"""The definition for the actions section of the toolbox."""

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection
from gaphor.diagram.diagramtools import PlacementTool
from gaphor.UML import diagramitems


def activity_config(new_item):
    subject = new_item.subject
    subject.name = f"New{type(subject).__name__}"
    if subject.activity:
        return

    diagram = new_item.diagram
    package = diagram.namespace

    activities = (
        [i for i in package.ownedClassifier if isinstance(i, UML.Activity)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.Activity) and e.package is None
        )
    )
    if activities:
        subject.activity = activities[0]
    else:
        activity = subject.model.create(UML.Activity)
        activity.name = "Activity"
        activity.package = package
        subject.activity = activity


def partition_config(new_item):
    activity_config(new_item)
    subject = new_item.subject
    new_item.partition = subject

    partition = subject.model.create(UML.ActivityPartition)
    partition.name = "NewActivityPartition"
    partition.activity = subject.activity
    new_item.partition = partition


actions = ToolSection(
    gettext("Actions"),
    (
        ToolDef(
            "toolbox-action",
            gettext("Action"),
            "gaphor-action-symbolic",
            "a",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.ActionItem,
                UML.Action,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-initial-node",
            gettext("Initial node"),
            "gaphor-initial-node-symbolic",
            "j",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.InitialNodeItem,
                UML.InitialNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-activity-final-node",
            gettext("Activity final node"),
            "gaphor-activity-final-node-symbolic",
            "f",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.ActivityFinalNodeItem,
                UML.ActivityFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-flow-final-node",
            gettext("Flow final node"),
            "gaphor-flow-final-node-symbolic",
            "w",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.FlowFinalNodeItem,
                UML.FlowFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-decision-node",
            gettext("Decision/merge node"),
            "gaphor-decision-node-symbolic",
            "g",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.DecisionNodeItem,
                UML.DecisionNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-fork-node",
            gettext("Fork/join node"),
            "gaphor-fork-node-symbolic",
            "<Shift>R",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.ForkNodeItem,
                UML.JoinNode,
                config_func=activity_config,
            ),
            handle_index=1,
        ),
        ToolDef(
            "toolbox-object-node",
            gettext("Object node"),
            "gaphor-object-node-symbolic",
            "<Shift>O",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.ObjectNodeItem,
                UML.ObjectNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-partition",
            gettext("Partition"),
            "gaphor-partition-symbolic",
            "<Shift>P",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.PartitionItem,
                UML.ActivityPartition,
                config_func=partition_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-flow",
            gettext("Control/object flow"),
            "gaphor-control-flow-symbolic",
            "<Shift>F",
            item_factory=PlacementTool.new_item_factory(diagramitems.FlowItem),
        ),
        ToolDef(
            "toolbox-send-signal-action",
            gettext("Send signal action"),
            "gaphor-send-signal-action-symbolic",
            None,
            item_factory=PlacementTool.new_item_factory(
                diagramitems.SendSignalActionItem,
                UML.SendSignalAction,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-accept-event-action",
            gettext("Accept event action"),
            "gaphor-accept-event-action-symbolic",
            None,
            item_factory=PlacementTool.new_item_factory(
                diagramitems.AcceptEventActionItem,
                UML.AcceptEventAction,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
    ),
)



from __future__ import annotations

import ast
from dataclasses import replace
from typing import Optional

import gaphas
from gaphas.aspect import ConnectionSink
from gaphas.aspect import Connector as ConnectorAspect
from gaphas.geometry import Rectangle, distance_rectangle_point

from gaphor.core.modeling.presentation import Presentation, S
from gaphor.core.styling import Style
from gaphor.diagram.shapes import combined_style
from gaphor.diagram.text import TextAlign, text_point_at_line


class Named:
    """Marker for any NamedElement presentations."""


class Classified(Named):
    """Marker for Classifier presentations."""


def from_package_str(item):
    """Display name space info when it is different, then diagram's or parent's
    namespace."""
    subject = item.subject
    canvas = item.canvas

    if not (subject and canvas):
        return False

    namespace = subject.namespace
    parent = canvas.get_parent(item)

    # if there is a parent (i.e. interaction)
    if parent and parent.subject and parent.subject.namespace is not namespace:
        return False

    return f"(from {namespace.name})" if namespace is not item.diagram.namespace else ""


def _get_sink(item, handle, target):
    assert item.canvas

    hpos = item.canvas.get_matrix_i2i(item, target).transform_point(*handle.pos)
    port = None
    dist = 10e6
    for p in target.ports():
        pos, d = p.glue(hpos)
        if not port or d < dist:
            port = p
            dist = d

    return ConnectionSink(target, port)


def postload_connect(item: gaphas.Item, handle: gaphas.Handle, target: gaphas.Item):
    """Helper function: when loading a model, handles should be connected as
    part of the `postload` step.

    This function finds a suitable spot on the `target` item to connect
    the handle to.
    """
    connector = ConnectorAspect(item, handle)
    sink = _get_sink(item, handle, target)
    connector.connect(sink)


# Note: the official documentation is using the terms "Shape" and "Edge" for element and line.


class ElementPresentation(Presentation[S], gaphas.Element):
    """Presentation for Gaphas Element (box-like) items.

    To create a shape (boxes, text), assign a shape to `self.shape`. If
    the shape can change, for example, because styling needs to change,
    implement the method `update_shapes()` and set self.shape there.
    """

    width: int
    height: int

    _port_sides = ("top", "right", "bottom", "left")

    def __init__(self, id=None, model=None, shape=None):
        super().__init__(id, model)
        self._shape = shape

    def port_side(self, port):
        return self._port_sides[self._ports.index(port)]

    def _set_shape(self, shape):
        self._shape = shape
        self.request_update()

    shape = property(lambda s: s._shape, _set_shape)

    def update_shapes(self, event=None):
        """Updating the shape configuration, e.g. when extra elements have to
        be drawn or when styling changes."""

    def pre_update(self, context):
        self.min_width, self.min_height = self.shape.size(context)

    def draw(self, context):
        self._shape.draw(
            context,
            Rectangle(0, 0, self.width, self.height),
        )

    def setup_canvas(self):
        super().setup_canvas()
        self.subscribe_all()
        # Invoke here, since we do not receive events, unless we're attached to a canvas
        self.update_shapes()

    def teardown_canvas(self):
        self.unsubscribe_all()
        super().teardown_canvas()

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        for prop in ("width", "height"):
            save_func(prop, getattr(self, prop))
        super().save(save_func)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name in ("width", "height"):
            setattr(self, name, ast.literal_eval(value))
        else:
            super().load(name, value)

    def postload(self):
        super().postload()
        self.update_shapes()


class LinePresentation(Presentation[S], gaphas.Line):
    def __init__(
        self,
        id=None,
        model=None,
        style: Style = {},
        shape_head=None,
        shape_middle=None,
        shape_tail=None,
    ):
        super().__init__(id, model)

        self.style = style
        self.shape_head = shape_head
        self.shape_middle = shape_middle
        self.shape_tail = shape_tail

        self.fuzziness = 2
        self._shape_head_rect = None
        self._shape_middle_rect = None
        self._shape_tail_rect = None

    canvas: Optional[gaphas.Canvas]

    head = property(lambda self: self._handles[0])
    tail = property(lambda self: self._handles[-1])

    def post_update(self, context):
        def shape_bounds(shape, align):
            if shape:
                size = shape.size(context)
                x, y = text_point_at_line(points, size, align)
                return Rectangle(x, y, *size)

        super().post_update(context)
        points = [h.pos for h in self.handles()]
        self._shape_head_rect = shape_bounds(self.shape_head, TextAlign.LEFT)
        self._shape_middle_rect = shape_bounds(self.shape_middle, TextAlign.CENTER)
        self._shape_tail_rect = shape_bounds(self.shape_tail, TextAlign.RIGHT)

    def point(self, pos):
        """Given a point (x, y) return the distance to the canvas item."""
        d0 = super().point(pos)
        ds = [
            distance_rectangle_point(shape, pos)
            for shape in (
                self._shape_head_rect,
                self._shape_middle_rect,
                self._shape_tail_rect,
            )
            if shape
        ]
        return min(d0, *ds) if ds else d0

    def draw(self, context):
        style = combined_style(context.style, self.style)
        context = replace(context, style=style)

        cr = context.cairo
        self.line_width = style["line-width"]
        cr.set_dash(style.get("dash-style", ()), 0)
        stroke = style["color"]
        if stroke:
            cr.set_source_rgba(*stroke)

        super().draw(context)

        for shape, rect in (
            (self.shape_head, self._shape_head_rect),
            (self.shape_middle, self._shape_middle_rect),
            (self.shape_tail, self._shape_tail_rect),
        ):
            if shape:
                shape.draw(context, rect)

    def setup_canvas(self):
        super().setup_canvas()
        self.subscribe_all()

    def teardown_canvas(self):
        self.unsubscribe_all()
        super().teardown_canvas()

    def save(self, save_func):
        def save_connection(name, handle):
            assert self.canvas
            c = self.canvas.get_connection(handle)
            if c:
                save_func(name, c.connected)

        super().save(save_func)
        save_func("matrix", tuple(self.matrix))
        for prop in ("orthogonal", "horizontal"):
            save_func(prop, getattr(self, prop))
        points = [tuple(map(float, h.pos)) for h in self.handles()]
        save_func("points", points)

        save_connection("head-connection", self.head)
        save_connection("tail-connection", self.tail)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name == "points":
            points = ast.literal_eval(value)
            for _ in range(len(points) - 2):
                h = self._create_handle((0, 0))
                self._handles.insert(1, h)
            for i, p in enumerate(points):
                self.handles()[i].pos = p

            # Update connection ports of the line. Only handles are saved
            # in Gaphor file therefore ports need to be recreated after
            # handles information is loaded.
            self._update_ports()

        elif name == "orthogonal":
            self._load_orthogonal = ast.literal_eval(value)
        elif name == "horizontal":
            self.horizontal = ast.literal_eval(value)
        elif name in ("head_connection", "head-connection"):
            self._load_head_connection = value
        elif name in ("tail_connection", "tail-connection"):
            self._load_tail_connection = value
        else:
            super().load(name, value)

    def postload(self):
        assert self.canvas

        if hasattr(self, "_load_orthogonal"):
            # Ensure there are enough handles
            if self._load_orthogonal and len(self._handles) < 3:
                p0 = self._handles[-1].pos
                self._handles.insert(1, self._create_handle(p0))
            self.orthogonal = self._load_orthogonal
            del self._load_orthogonal

        if hasattr(self, "_load_head_connection"):
            postload_connect(self, self.head, self._load_head_connection)
            del self._load_head_connection

        if hasattr(self, "_load_tail_connection"):
            postload_connect(self, self.tail, self._load_tail_connection)
            del self._load_tail_connection

        super().postload()



"""Test messages."""

from gaphor import UML
from gaphor.diagram.grouping import Group
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.message import MessageItem


def test_message_persistence(diagram, element_factory, saver, loader):
    diagram.create(MessageItem, subject=element_factory.create(UML.Message))

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(UML.Diagram))
    item = new_diagram.canvas.select(MessageItem)[0]

    assert item


def test_group_message_item_without_subject(diagram, element_factory):
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )
    message = diagram.create(MessageItem)

    Group(interaction, message).group()

    assert message.subject is None


def test_group_message_item_with_subject(diagram, element_factory):
    interaction = diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )
    message = diagram.create(MessageItem, subject=element_factory.create(UML.Message))

    Group(interaction, message).group()

    assert message.subject
    assert message.subject.interaction is interaction.subject



"""The definition for the states section of the toolbox."""

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection
from gaphor.diagram.diagramtools import PlacementTool
from gaphor.UML import diagramitems


def initial_pseudostate_config(new_item):
    new_item.subject.kind = "initial"
    state_machine_config(new_item)


def history_pseudostate_config(new_item):
    new_item.subject.kind = "shallowHistory"
    state_machine_config(new_item)


def state_machine_config(new_item):
    subject = new_item.subject
    subject.name = f"New{type(subject).__name__}"
    if subject.container:
        return

    diagram = new_item.diagram
    package = diagram.namespace

    state_machines = (
        [i for i in package.ownedClassifier if isinstance(i, UML.StateMachine)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.StateMachine) and e.package is None
        )
    )

    if state_machines:
        state_machine = state_machines[0]
    else:
        state_machine = subject.model.create(UML.StateMachine)
        state_machine.name = "StateMachine"
        state_machine.package = package

    if state_machine.region:
        region = state_machine.region[0]
    else:
        region = subject.model.create(UML.Region)
        region.name = "DefaultRegion"
        region.stateMachine = state_machine

    subject.container = region


states = ToolSection(
    gettext("States"),
    (
        ToolDef(
            "toolbox-state",
            gettext("State"),
            "gaphor-state-symbolic",
            "s",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.StateItem, UML.State, config_func=state_machine_config
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-initial-pseudostate",
            gettext("Initial Pseudostate"),
            "gaphor-initial-pseudostate-symbolic",
            "<Shift>S",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                initial_pseudostate_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-final-state",
            gettext("Final State"),
            "gaphor-final-state-symbolic",
            "x",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.FinalStateItem,
                UML.FinalState,
                config_func=state_machine_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-history-pseudostate",
            gettext("History Pseudostate"),
            "gaphor-pseudostate-symbolic",
            "q",
            item_factory=PlacementTool.new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                history_pseudostate_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-transition",
            gettext("Transition"),
            "gaphor-transition-symbolic",
            "<Shift>T",
            item_factory=PlacementTool.new_item_factory(diagramitems.TransitionItem),
        ),
    ),
)



import pytest

from gaphor import UML
from gaphor.ui.diagrampage import tooliter
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.toolbox import uml_toolbox_actions


@pytest.fixture
def interaction(diagram, element_factory):
    return diagram.create(
        InteractionItem, subject=element_factory.create(UML.Interaction)
    )


@pytest.fixture
def lifeline_factory():
    return next(
        t for t in tooliter(uml_toolbox_actions) if t.id == "toolbox-lifeline"
    ).item_factory


def test_create_lifeline_on_diagram_should_create_an_interaction(
    diagram, lifeline_factory
):
    lifeline = lifeline_factory(diagram)

    assert lifeline.subject.interaction


def test_create_lifeline_on_diagram_should_use_existing_interaction(
    diagram, lifeline_factory, element_factory
):
    interaction = element_factory.create(UML.Interaction)
    lifeline = lifeline_factory(diagram)

    assert lifeline.subject.interaction is interaction


def test_create_lifeline_on_diagram_in_package_should_create_an_interaction(
    diagram, element_factory, lifeline_factory
):
    package = element_factory.create(UML.Package)
    diagram.package = package
    lifeline = lifeline_factory(diagram)

    assert lifeline.subject.interaction
    assert lifeline.subject.interaction.package is package


def test_create_lifeline_over_interaction_uses_that_interaction(
    diagram, interaction, lifeline_factory
):
    lifeline = lifeline_factory(diagram, parent=interaction)

    assert lifeline.subject.interaction is interaction.subject



import pytest

from gaphor.ui.event import Notification
from gaphor.ui.mainwindow import new_builder
from gaphor.ui.notification import InAppNotifier


@pytest.fixture
def builder():
    return new_builder()


@pytest.fixture
def in_app_notifier(builder):
    return InAppNotifier(builder)


def test_notifier_has_widgets(in_app_notifier):
    assert in_app_notifier.revealer
    assert in_app_notifier.message_label


def test_notifier_can_handle_message(in_app_notifier):
    in_app_notifier.handle(Notification("a message"))

    assert in_app_notifier.revealer.get_reveal_child()



import pytest

from gaphor import UML
from gaphor.ui.diagrampage import tooliter
from gaphor.UML.toolbox import uml_toolbox_actions

activity_node_names = [
    "action",
    "initial-node",
    "activity-final-node",
    "flow-final-node",
    "decision-node",
    "fork-node",
    "object-node",
    "send-signal-action",
    "accept-event-action",
]


@pytest.fixture
def item_factory(request):
    return next(
        t for t in tooliter(uml_toolbox_actions) if t.id == f"toolbox-{request.param}"
    ).item_factory


@pytest.mark.parametrize("item_factory", activity_node_names, indirect=True)
def test_create_action_should_create_an_activity(diagram, item_factory):
    action = item_factory(diagram)

    assert action.subject.activity
    assert action.subject.owner is action.subject.activity


@pytest.mark.parametrize("item_factory", activity_node_names, indirect=True)
def test_create_action_should_add_to_existing_activity(
    diagram, item_factory, element_factory
):
    activity = element_factory.create(UML.Activity)
    action = item_factory(diagram)

    assert action.subject.activity is activity


@pytest.mark.parametrize("item_factory", activity_node_names, indirect=True)
def test_create_action_should_add_to_existing_activity_in_package(
    diagram, item_factory, element_factory
):
    package = element_factory.create(UML.Package)
    diagram.package = package
    activity = element_factory.create(UML.Activity)
    activity.package = package
    action = item_factory(diagram)

    assert action.subject.activity is activity



"""Classes related adapter connection tests."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect, get_connected
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.classes.generalization import GeneralizationItem
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.usecases.actor import ActorItem


def test_dependency_glue(create):
    """Test dependency glue to two actor items."""
    actor1 = create(ActorItem, UML.Actor)
    actor2 = create(ActorItem, UML.Actor)
    dep = create(DependencyItem)

    glued = allow(dep, dep.head, actor1)
    assert glued

    connect(dep, dep.head, actor1)

    glued = allow(dep, dep.tail, actor2)
    assert glued


def test_dependency_connect(create, element_factory):
    """Test dependency connecting to two actor items."""
    actor1 = create(ActorItem, UML.Actor)
    actor2 = create(ActorItem, UML.Actor)
    dep = create(DependencyItem)

    connect(dep, dep.head, actor1)
    connect(dep, dep.tail, actor2)

    assert dep.subject is not None
    assert isinstance(dep.subject, UML.Dependency)
    assert dep.subject in element_factory.select()

    hct = get_connected(dep, dep.head)
    tct = get_connected(dep, dep.tail)
    assert hct is actor1
    assert tct is actor2

    assert actor1.subject is dep.subject.supplier
    assert actor2.subject is dep.subject.client


def test_dependency_reconnection(create):
    """Test dependency reconnection."""
    a1 = create(ActorItem, UML.Actor)
    a2 = create(ActorItem, UML.Actor)
    a3 = create(ActorItem, UML.Actor)
    dep = create(DependencyItem)

    # connect: a1 -> a2
    connect(dep, dep.head, a1)
    connect(dep, dep.tail, a2)

    d = dep.subject

    # reconnect: a1 -> a3
    connect(dep, dep.tail, a3)

    assert d is dep.subject
    assert a1.subject is dep.subject.supplier
    assert a3.subject is dep.subject.client
    assert a2.subject is not dep.subject.client, dep.subject.client


def test_dependency_disconnect(create, element_factory):
    actor1 = create(ActorItem, UML.Actor)
    actor2 = create(ActorItem, UML.Actor)
    dep = create(DependencyItem)

    connect(dep, dep.head, actor1)
    connect(dep, dep.tail, actor2)

    dep_subj = dep.subject
    disconnect(dep, dep.tail)

    assert dep.subject is None
    assert get_connected(dep, dep.tail) is None
    assert dep_subj not in element_factory.select()
    assert dep_subj not in actor1.subject.supplierDependency
    assert dep_subj not in actor2.subject.clientDependency


def test_dependency_reconnect(create):
    """Test dependency reconnection using two actor items."""
    actor1 = create(ActorItem, UML.Actor)
    actor2 = create(ActorItem, UML.Actor)
    dep = create(DependencyItem)

    connect(dep, dep.head, actor1)
    connect(dep, dep.tail, actor2)

    dep_subj = dep.subject
    disconnect(dep, dep.tail)

    # reconnect
    connect(dep, dep.tail, actor2)

    assert dep.subject is not None
    assert dep.subject is not dep_subj  # the old subject has been deleted
    assert dep.subject in actor1.subject.supplierDependency
    assert dep.subject in actor2.subject.clientDependency
    # TODO: test with interface (usage) and component (realization)
    # TODO: test with multiple diagrams (should reuse existing relationships first)


def test_multiple_dependencies(create, element_factory):
    """Test multiple dependencies.

    Dependency should appear in a new diagram, bound on a new dependency
    item.
    """
    actoritem1 = create(ActorItem, UML.Actor)
    actoritem2 = create(ActorItem, UML.Actor)
    actor1 = actoritem1.subject
    actor2 = actoritem2.subject
    dep = create(DependencyItem)

    connect(dep, dep.head, actoritem1)
    connect(dep, dep.tail, actoritem2)

    assert dep.subject
    assert 1 == len(actor1.supplierDependency)
    assert actor1.supplierDependency[0] is dep.subject
    assert 1 == len(actor2.clientDependency)
    assert actor2.clientDependency[0] is dep.subject

    # Do the same thing, but now on a new diagram:

    diagram2 = element_factory.create(UML.Diagram)
    actoritem3 = diagram2.create(ActorItem, subject=actor1)
    actoritem4 = diagram2.create(ActorItem, subject=actor2)
    dep2 = diagram2.create(DependencyItem)

    connect(dep2, dep2.head, actoritem3)
    cinfo = diagram2.canvas.get_connection(dep2.head)
    assert cinfo is not None
    assert cinfo.connected is actoritem3
    connect(dep2, dep2.tail, actoritem4)
    assert dep2.subject is not None
    assert 1 == len(actor1.supplierDependency)
    assert actor1.supplierDependency[0] is dep.subject
    assert 1 == len(actor2.clientDependency)
    assert actor2.clientDependency[0] is dep.subject

    assert dep.subject is dep2.subject


def test_dependency_type_auto(create, element_factory):
    """Test dependency type automatic determination."""
    cls = create(ClassItem, UML.Class)
    iface = create(InterfaceItem, UML.Interface)
    dep = create(DependencyItem)

    assert dep.auto_dependency

    connect(dep, dep.tail, cls)  # connect client
    connect(dep, dep.head, iface)  # connect supplier

    assert dep.subject is not None
    assert isinstance(dep.subject, UML.Usage), dep.subject
    assert dep.subject in element_factory.select()


def test_generalization_glue(create):
    """Test generalization item gluing using two classes."""

    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    glued = allow(gen, gen.tail, c1)
    assert glued

    connect(gen, gen.tail, c1)
    assert get_connected(gen, gen.tail) is c1
    assert gen.subject is None

    glued = allow(gen, gen.head, c2)
    assert glued


def test_generalization_connection(create):
    """Test generalization item connection using two classes."""
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.tail, c1)
    assert get_connected(gen, gen.tail) is c1

    connect(gen, gen.head, c2)
    assert gen.subject is not None
    assert gen.subject.general is c2.subject
    assert gen.subject.specific is c1.subject


def test_generalization_reconnection(create, element_factory):
    """Test generalization item connection using two classes.

    On reconnection a new Generalization is created.
    """
    gen = create(GeneralizationItem)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)

    connect(gen, gen.tail, c1)
    assert get_connected(gen, gen.tail) is c1

    connect(gen, gen.head, c2)
    assert gen.subject is not None
    assert gen.subject.general is c2.subject
    assert gen.subject.specific is c1.subject

    # Now do the same on a new diagram:
    diagram2 = element_factory.create(UML.Diagram)
    c3 = diagram2.create(ClassItem, subject=c1.subject)
    c4 = diagram2.create(ClassItem, subject=c2.subject)
    gen2 = diagram2.create(GeneralizationItem)

    connect(gen2, gen2.head, c3)
    cinfo = diagram2.canvas.get_connection(gen2.head)
    assert cinfo is not None
    assert cinfo.connected is c3

    connect(gen2, gen2.tail, c4)
    assert gen.subject is not gen2.subject
    assert len(c1.subject.generalization) == 1
    assert c1.subject.generalization[0] is gen.subject
    # self.assertEqual(1, len(actor2.clientDependency))
    # self.assertTrue(actor2.clientDependency[0] is dep.subject)


def test_generalization_reconnection2(create):
    """Test reconnection of generalization."""
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    c3 = create(ClassItem, UML.Class)
    gen = create(GeneralizationItem)

    # connect: c1 -> c2
    connect(gen, gen.head, c1)
    connect(gen, gen.tail, c2)

    s = gen.subject

    # reconnect: c2 -> c3
    connect(gen, gen.tail, c3)

    assert s is gen.subject
    assert c1.subject is gen.subject.general
    assert c3.subject is gen.subject.specific
    assert c2.subject is not gen.subject.specific



import pytest

from gaphor import UML
from gaphor.diagram.diagramtools import PlacementTool
from gaphor.UML.actions.actionstoolbox import partition_config
from gaphor.UML.actions.partition import PartitionItem


@pytest.fixture
def partition_item_factory():
    return PlacementTool.new_item_factory(
        PartitionItem,
        UML.ActivityPartition,
        config_func=partition_config,
    )


@pytest.fixture
def new_partition(partition_item_factory, diagram):
    return partition_item_factory(diagram)


def test_partition_placement_adds_two_partitions(new_partition: PartitionItem):

    assert new_partition.subject in new_partition.partition
    assert len(new_partition.partition) == 2



import pytest

from gaphor import UML
from gaphor.ui.diagrampage import tooliter
from gaphor.UML.toolbox import uml_toolbox_actions

state_node_names = [
    "state",
    "initial-pseudostate",
    "final-state",
    "history-pseudostate",
]


@pytest.fixture
def item_factory(request):
    return next(
        t for t in tooliter(uml_toolbox_actions) if t.id == f"toolbox-{request.param}"
    ).item_factory


@pytest.mark.parametrize("item_factory", state_node_names, indirect=True)
def test_create_state_on_diagram_should_create_a_state_machine_with_region(
    diagram, item_factory
):
    state = item_factory(diagram)

    assert state.subject.container
    assert state.subject.container.stateMachine


@pytest.mark.parametrize("item_factory", state_node_names, indirect=True)
def test_create_state_should_add_to_existing_state_machine(
    diagram, item_factory, element_factory
):
    state_machine = element_factory.create(UML.StateMachine)
    state = item_factory(diagram)
    region = state.subject.container

    assert region
    assert region.stateMachine is state_machine


@pytest.mark.parametrize("item_factory", state_node_names, indirect=True)
def test_create_state_should_add_to_existing_state_machine_and_region(
    diagram, item_factory, element_factory
):
    state_machine = element_factory.create(UML.StateMachine)
    region = element_factory.create(UML.Region)
    region.stateMachine = state_machine

    state = item_factory(diagram)

    assert state.subject.container is region


@pytest.mark.parametrize("item_factory", state_node_names, indirect=True)
def test_create_state_should_add_to_existing_state_machine_and_region_in_package(
    diagram, item_factory, element_factory
):
    state_machine = element_factory.create(UML.StateMachine)
    region = element_factory.create(UML.Region)
    region.stateMachine = state_machine

    package = element_factory.create(UML.Package)
    diagram.package = package
    state_machine.package = package

    state = item_factory(diagram)

    assert state.subject.container is region


@pytest.mark.parametrize("item_factory", state_node_names, indirect=True)
def test_create_state_should_add_to_existing_state_machine_in_package(
    diagram, item_factory, element_factory
):
    state_machine = element_factory.create(UML.StateMachine)

    package = element_factory.create(UML.Package)
    diagram.package = package
    state_machine.package = package

    state = item_factory(diagram)

    assert state.subject.container.stateMachine is state_machine



import pytest

from gaphor.diagram.tests.fixtures import diagram, element_factory, event_manager


@pytest.fixture
def create(diagram, element_factory):
    def _create(item_class, element_class=None):
        return diagram.create(
            item_class,
            subject=(element_factory.create(element_class) if element_class else None),
        )

    return _create



import pytest

from gaphor.storage.parser import element
from gaphor.storage.storage import load_elements


@pytest.fixture
def loader(element_factory, modeling_language):
    def _loader(*parsed_elements):
        parsed_data = {p.id: p for p in parsed_elements}
        load_elements(parsed_data, element_factory, modeling_language)
        return element_factory.lselect()

    return _loader


def test_owned_comment_to_comment_upgrade(loader):
    c = element(id="1", type="Comment")
    c.references["annotatedElement"] = ["2"]
    e = element(id="2", type="Element")
    e.references["ownedComment"] = ["1"]

    comment, elem = loader(c, e)

    assert elem in comment.annotatedElement
    assert comment in elem.comment



"""Tests for grouping functionality in Gaphor."""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.actions import ActionItem, PartitionItem


class PartitionGroupTestCase(TestCase):
    def test_no_subpartition_when_nodes_in(self):
        """Test adding subpartition when nodes added."""
        p = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        a1 = self.create(ActionItem, UML.Action)
        p1 = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        p2 = self.create(PartitionItem, subject_cls=UML.ActivityPartition)

        self.group(p, p1)
        self.group(p1, a1)
        assert not self.can_group(p1, p2)

    def test_action_grouping(self):
        """Test adding action to partition."""
        p1 = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        p2 = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        assert self.can_group(p1, a1)
        self.group(p1, a1)
        assert len(p1.subject.node) == 1

        self.group(p1, p2)
        self.group(p2, a1)

        assert self.can_group(p2, a1)
        assert len(p2.subject.node) == 1
        self.group(p2, a2)
        assert len(p2.subject.node) == 2

    def test_ungrouping(self):
        """Test action and partition removal."""
        p = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        self.group(p, a1)
        self.group(p, a2)

        self.ungroup(p, a1)
        assert len(p.subject.node) == 1
        self.ungroup(p, a2)
        assert len(p.subject.node) == 0

    def test_ungrouping_with_actions(self):
        """Test partition with actions removal."""
        p = self.create(PartitionItem, subject_cls=UML.ActivityPartition)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        self.group(p, a1)
        self.group(p, a2)

        partition = p.subject
        assert len(partition.node) == 2, partition.node
        assert len(p.canvas.get_children(p)) == 2, p.canvas.get_children(p)

        self.ungroup(p, a1)
        self.ungroup(p, a2)

        assert 0 == len(partition.node)
        assert len(p.canvas.get_children(p)) == 0
        assert len(partition.node) == 0
